# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import mock
import unittest

from .. import AuthProxy
from .. import HelpScout
from ..apis import __all__ as all_apis


class TestHelpScout(unittest.TestCase):

    API_KEY = 'test key'

    def setUp(self):
        super(TestHelpScout, self).setUp()
        self.hs = HelpScout(self.API_KEY)

    def test_init_session(self):
        """It should build a session with the proper authentication."""
        self.assertEqual(self.hs.session.auth.username, self.API_KEY)

    def test_load_apis(self):
        """It should load all available APIs."""
        self.assertEqual(len(self.hs.__apis__), len(all_apis))

    def test_api_instance_attributes(self):
        """It should properly set all of the API instance attributes."""
        for api_name, api in self.hs.__apis__.items():
            self.assertEqual(getattr(self.hs, api_name), api)

    def test_api_auth_proxy(self):
        """It should wrap the APIs in an AuthProxy."""
        for api in self.hs.__apis__.values():
            self.assertIsInstance(api, AuthProxy)

    @mock.patch('helpscout.WebHook')
    def test_web_hook_init(self, WebHook):
        """It should initialize a web hook using the API key."""
        self.hs.web_hook()
        WebHook.assert_called_once_with(self.API_KEY)

    @mock.patch('helpscout.WebHook')
    def test_web_hook_cache(self, WebHook):
        """It should use the cached web hook the second time around."""
        self.hs.web_hook()
        self.hs.web_hook()
        WebHook.assert_called_once_with(self.API_KEY)

    @mock.patch('helpscout.WebHook')
    def test_web_hook_pass_through(self, WebHook):
        """It should pass args and kwargs through to receive method."""
        args = [1, 2, 3]
        kwargs = {'test1': 'derp', 'test2': True}
        self.hs.web_hook(*args, **kwargs)
        WebHook().receive.assert_called_once_with(*args, **kwargs)

    @mock.patch('helpscout.WebHook')
    def test_web_hook_return(self, WebHook):
        """It should return the receive method on a new web hook."""
        res = self.hs.web_hook()
        self.assertEqual(res, WebHook().receive())



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Chat(BaseModel):
    """This represents a chat (IM) address."""

    value = properties.String(
        'Value',
        required=True,
    )
    type = properties.StringChoice(
        'Type',
        choices=['aim',
                 'gtalk',
                 'icq',
                 'xmpp',
                 'msn',
                 'skype',
                 'yahoo',
                 'qq',
                 'other',
                 ],
        default='other',
        required=True,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from requests import Session
from requests.auth import HTTPBasicAuth

from .auth_proxy import AuthProxy
from .base_api import BaseApi
from .base_model import BaseModel
from .domain import Domain
from .web_hook import WebHook

from . import exceptions


class HelpScout(object):
    """This object is the primary point of interaction with HelpScout.

    Properties will be set on this ``HelpScout`` instance that will mirror
    the class names of APIs in the ``helpscout.api`` module.

    These API classes are naive of authentication, so the actual properties
    set will be using the ``AuthProxy`` class, which will transparently
    inject authentication into the API requests while still allowing for a
    naive API object.

    This allows for the ``HelpScout`` instance to act as a container for
    all of the authenticated API objects.

    Examples::

        from helpscout import HelpScout
        hs = HelpScout('api_key')
        for customer in hs.Customers.list():
            print(customer)

    Attributes:
        Conversations (helpscout.api.conversations.Conversations):
            Conversations API endpoint.
        Customers (helpscout.api.customers.Customers): Customers API
            endpoint.
        Mailboxes (helpscout.api.mailboxes.Mailboxes): Mailboxes API endpoint.
        Tags (helpscout.api.tags.Tags): Tags API endpoint.
        Teams (helpscout.api.teams.Teams): Teams API endpoint.
        Users (helpscout.api.users.Users): Users API endpoint.
        __apis__ (dict): References to all available APIs, keyed by class
            name.
        web_hook (WebHook): A usable web hook parser object; lazy loaded when
            actually used.
    """

    __apis__ = {}
    _web_hook = None

    def __init__(self, api_key):
        """Initialize a new HelpScout client.

        Args:
            api_key (str): The API key to use for this session.
        """
        self.session = Session()
        self.session.auth = HTTPBasicAuth(api_key, 'NoPassBecauseKey!')
        self._load_apis()

    def web_hook(self, *args, **kwargs):
        """Represents the web hook parser. Is lazy loaded when needed.

        Returns:
            helpscout.web_hook.WebHookEvent: The event that was received.
        """
        if self._web_hook is None:
            self._web_hook = WebHook(self.session.auth.username)
        return self._web_hook.receive(*args, **kwargs)

    def _load_apis(self):
        """Find available APIs and set instances property auth proxies."""
        helpscout = __import__('helpscout.apis')
        for class_name in helpscout.apis.__all__:
            if not class_name.startswith('_'):
                cls = getattr(helpscout.apis, class_name)
                api = AuthProxy(self.session, cls)
                setattr(self, class_name, api)
                self.__apis__[class_name] = api


__all__ = [
    'AuthProxy',
    'BaseApi',
    'BaseModel',
    'Domain',
    'exceptions',
    'HelpScout',
]



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Attachment(BaseModel):

    hash = properties.String(
        'Unique hash.',
        required=True,
    )
    mime_type = properties.String(
        'Mime Type',
    )
    filename = properties.String(
        'File Name',
        required=True,
    )
    size = properties.Integer(
        'Size of the attachment in bytes.',
    )
    width = properties.Integer(
        'Image width',
    )
    height = properties.Integer(
        'Image height',
    )
    url = properties.String(
        'Public-facing url where attachment can be downloaded.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import hmac
import json
import properties

from base64 import b64encode
from hashlib import sha1
from six import string_types

from .web_hook_event import WebHookEvent

from ..exceptions import HelpScoutSecurityException


class WebHook(properties.HasProperties):
    """``WebHook`` provides the ability to easily process web hook events."""

    api_key = properties.String(
        'API Key that should be used to validate requests against.',
        required=True,
    )

    def receive(self, event_type, signature, data_str):
        """Receive a web hook for the event and signature.

        Args:
            event_type (str): Name of the event that was received (from the
                request ``X-HelpScout-Event`` header).
            signature (str): The signature that was received, which serves as
                authentication (from the request ``X-HelpScout-Signature``
                header).
            data_str (str): The raw data that was posted by HelpScout
                to the web hook. This must be the raw string, because if it
                is parsed with JSON it will lose its ordering and not pass
                signature validation.

        Raises:
            HelpScoutSecurityException: If an invalid signature is provided,
                and ``raise_if_invalid`` is ``True``.

        Returns:
            WebHookEvent: The authenticated web hook request.
        """

        if not self.validate_signature(signature, data_str):
            raise HelpScoutSecurityException(
                'The signature provided by this request was invalid.',
            )

        return WebHookEvent(
            event_type=event_type,
            record=json.loads(data_str),
        )

    def validate_signature(self, signature, data, encoding='utf8'):
        """Validate the signature for the provided data.

        Args:
            signature (str | bytes | bytearray): Signature that was provided
                for the request.
            data (str | bytes | bytearray): Data string to validate against
                the signature.
            encoding (str, optional): If a string was provided for ``data`` or
                ``signature``, this is the character encoding.

        Returns:
            bool: Whether the signature is valid for the provided data.
        """

        if isinstance(data, string_types):
            data = bytearray(data, encoding)
        if isinstance(signature, string_types):
            signature = bytearray(signature, encoding)

        api_key = bytearray(self.api_key, 'utf8')
        hash = hmac.new(api_key, data, sha1)
        encoded = b64encode(hash.digest())

        return encoded.strip() == signature.strip()



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Option(BaseModel):
    """This represents option for a custom field of type ``DROPDOWN``."""

    label = properties.String(
        'Label of the option.',
        required=True,
    )
    order = properties.Integer(
        'Relative order of the custom field. Can be ``null`` or a number '
        'between ``0`` and ``255``.',
        min=0,
        max=255,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import unittest

from ..models import Customer
from ..web_hook.web_hook_event import WebHookEvent


class TestWebHookEvent(unittest.TestCase):

    def setUp(self):
        super(TestWebHookEvent, self).setUp()
        self.event_type = 'customer.something.happened'
        self.data = {
            'gender': 'male',
            'first_name': 'Test',
            'last_name': 'Dude',
            'type': 'customer',
        }
        self.api_data = {
            'gender': self.data['gender'],
            'firstName': self.data['first_name'],
            'lastName': self.data['last_name'],
            'type': self.data['type'],
        }

    def _validate_record(self, record):
        self.assertIsInstance(record.record, Customer)
        for key, val in self.data.items():
            self.assertEqual(getattr(record.record, key), val)

    def test_record_from_data(self):
        """It should create a correct record from the API data."""
        self._validate_record(
            WebHookEvent(event_type=self.event_type, record=self.api_data),
        )

    def test_record_from_record(self):
        """It should pass through a pre-existing record."""
        record = Customer(**self.data)
        self._validate_record(
            WebHookEvent(event_type=self.event_type, record=record),
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import json
import unittest

from ..exceptions import HelpScoutSecurityException
from ..web_hook import WebHook
from ..web_hook.web_hook_event import WebHookEvent


class TestWebHook(unittest.TestCase):

    def setUp(self):
        super(TestWebHook, self).setUp()
        self.data_str = '{"firstName":"Jackie","lastName":"Chan",' \
                        '"email":"jackie.chan@somewhere.com",' \
                        '"gender":"male"}'
        self.data = json.loads(self.data_str)
        self.signature = '\n2iFmnzC8SCNVF/iNiMnSe19yceU=\n'
        self.bad_signature = '4+kjl2zB5OwZZjHznSTkD1/J208='
        self.hook = WebHook(api_key='your secret key')

    def test_validate_signature_valid(self):
        """It should return True for valid signature."""
        self.assertTrue(
            self.hook.validate_signature(self.signature, self.data_str),
        )

    def test_validate_signature_invalid(self):
        """It should return False for invalid signature."""
        self.assertFalse(
            self.hook.validate_signature(self.bad_signature, self.data_str),
        )

    def test_receive_str_invalid(self):
        """It should raise on invalid signature from data string."""
        with self.assertRaises(HelpScoutSecurityException):
            self.hook.receive(None, self.signature, '{"nope": "1"}')

    def test_receive_returns_event(self):
        """Valid receive should return instantiated web hook event."""
        self.assertIsInstance(
            self.hook.receive(
                'customer.created', self.signature, self.data_str,
            ),
            WebHookEvent,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import unittest

from .. import AuthProxy


class EndTestException(Exception):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        super(EndTestException, self).__init__()


class ApiClass(object):

    @classmethod
    def proxied(cls, session=None):
        raise EndTestException(session=session)

    @classmethod
    def not_proxied(cls, *args, **kwargs):
        raise Exception()


class TestAuthProxy(unittest.TestCase):

    SESSION = 'session'

    def new_proxy(self, session=SESSION, proxy_class=ApiClass):
        return AuthProxy(session, proxy_class)

    def test_init_bad_api_class(self):
        """It should assert API class is a class."""
        with self.assertRaises(AssertionError):
            AuthProxy(None, 'not a class')

    def test_init_sets_session(self):
        """It should set the session instance var."""
        proxy = self.new_proxy()
        self.assertEqual(proxy.session, self.SESSION)

    def test_init_sets_proxy_class(self):
        """It should set the proxy_class instance var."""
        proxy = self.new_proxy()
        self.assertEqual(proxy.proxy_class, ApiClass)

    def test_getattr_injects_session(self):
        """It should inject the session into the proxied method."""
        proxy_method = self.new_proxy().proxied
        try:
            proxy_method()
        except EndTestException as e:
            self.assertEqual(e.kwargs['session'], self.SESSION)

    def test_getattr_whitelist(self):
        """It should not proxy methods that are in the whitelist."""
        proxy = self.new_proxy()
        proxy.METHOD_NO_PROXY.append('not_proxied')
        with self.assertRaises(AttributeError):
            proxy.not_proxied()



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import mock
import unittest

from .. import BaseApi
from .. import BaseModel
from ..domain import Domain
from ..request_paginator import RequestPaginator


PAGINATOR = 'helpscout.base_api.RequestPaginator'


class TestApi(BaseApi):
    __object__ = BaseModel


class TestBaseApi(unittest.TestCase):

    ENDPOINT = '/endpoint'
    DATA = {'test': 1234}
    REQUEST_TYPE = RequestPaginator.POST

    def new_api(self, endpoint=ENDPOINT, data=DATA, request_type=REQUEST_TYPE,
                singleton=False, session=None, out_type=None):
        return TestApi(
            endpoint, data, request_type, singleton, session, out_type,
        )

    def test_new(self):
        """It should return a new TestApi instance."""
        self.assertIsInstance(self.new_api(), TestApi)

    def test_new_paginator(self):
        """It should return a new API object with a paginator."""
        self.assertIsInstance(self.new_api().paginator, RequestPaginator)

    def test_new_paginator_create(self):
        """It should create the paginator with the proper args."""
        session = 'session'
        with mock.patch.object(RequestPaginator, '__init__') as init:
            init.return_value = None
            self.new_api(session=session)
            init.assert_called_once_with(
                endpoint='https://api.helpscout.net/v1%s' % self.ENDPOINT,
                data=self.DATA,
                output_type=TestApi.__object__.from_api,
                request_type=self.REQUEST_TYPE,
                session=session,
            )

    @mock.patch(PAGINATOR)
    def test_new_paginator_singleton(self, paginator):
        """It should return the record if singleton and found."""
        paginator().call.return_value = [{'id': 9876}]
        res = self.new_api(singleton=True)
        self.assertIsInstance(res, BaseModel)
        self.assertEqual(res.id, 9876)

    @mock.patch(PAGINATOR)
    def test_new_paginator_singleton_none(self, paginator):
        """It should return None if singleton and not found."""
        paginator().call.return_value = []
        res = self.new_api(singleton=True)
        self.assertIs(res, None)

    @mock.patch(PAGINATOR)
    def test_base_api_iterates_paginator(self, paginator):
        """It should pass iteration to the paginator."""
        expect = [BaseModel(id=1), BaseModel(id=2)]
        paginator().call.return_value = expect
        for idx, value in enumerate(self.new_api()):
            self.assertEqual(value, expect[idx])

    @mock.patch(PAGINATOR)
    def test_new_paginator_singleton_out_type(self, paginator):
        """It should return an object of the correct type if defined."""
        paginator().call.return_value = [{'id': 9876}]
        res = self.new_api(singleton=True, out_type=mock.MagicMock())
        self.assertIsInstance(res, mock.MagicMock)

    def test_get_search_domain_domain(self):
        """It should return the input if it is a `Domain` obj."""
        expect = Domain()
        self.assertEqual(BaseApi.get_search_domain(expect), expect)

    def test_get_search_domain_tuple(self):
        """It should return the input if it is a `Domain` obj."""
        expect = [('state', True)]
        with mock.patch.object(Domain, 'from_tuple') as from_tuple:
            res = BaseApi.get_search_domain(expect)
            from_tuple.assert_called_once_with(expect)
            self.assertEqual(res, from_tuple())

    def test_new_object(self):
        """It should return the proper object."""
        expect = 123
        res = BaseApi.new_object({'id': expect})
        self.assertIsInstance(res, BaseModel)
        self.assertEqual(res.id, expect)



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class BaseConversation(BaseModel):
    """This represents a basic conversation, meant to be subclassed."""

    number = properties.Integer(
        'The conversation number displayed in the UI. This number can be used '
        'in combination with the id to construct a URI to the conversation on '
        'the Help Scout website. Example: '
        '``https://secure.helpscout.net/conversation/<id>/<number>/``',
        required=True,
    )
    subject = properties.String(
        'The subject of the conversation.',
        required=True,
    )
    status = properties.StringChoice(
        'Status of the conversation.',
        choices=['action', 'pending', 'closed', 'spam'],
        default='pending',
        required=True,
    )
    thread_count = properties.Integer(
        'This count represents the number of published threads found on the '
        'conversation (it does not include line items, drafts or threads held '
        'for review by Traffic Cop).',
        required=True,
    )
    preview = properties.String(
        'Conversation preview.',
    )
    modified_at = properties.DateTime(
        'UTC time when a user last modified this conversation.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..models.folder import Folder
from ..models.mailbox import Mailbox


class Mailboxes(BaseApi):
    """This represents the ``Mailboxes`` Endpoint.

    The following aspects are implemented:

    * `List Mailboxes
      <http://developer.helpscout.net/help-desk-api/mailboxes/list/>`_
      (:func:`helpscout.apis.mailboxes.Mailboxes.list`)
    * `Get Mailboxes
      <http://developer.helpscout.net/help-desk-api/mailboxes/get/>`_
      (:func:`helpscout.apis.mailboxes.Mailboxes.get`)
    * `Get Folders
      <http://developer.helpscout.net/help-desk-api/mailboxes/folders/>`_
      (:func:`helpscout.apis.mailboxes.Mailboxes.get_folders`)
    """

    __object__ = Mailbox

    @classmethod
    def get(cls, session, mailbox_id):
        """Return a specific mailbox.

        Args:
            session (requests.Session): Authenticated requests Session.
            mailbox_id (int): The ID of the mailbox to get.

        Returns:
            Mailbox: A mailbox singleton, if existing.
            None: If no match for the ID.
        """
        return cls(
            '/mailboxes/%d.json' % mailbox_id,
            singleton=True,
            session=session,
        )

    @classmethod
    def list(cls, session):
        """List the mailboxes.

        Args:
            session (requests.Session): Authenticated requests Session.

        Returns:
            RequestPaginator of Mailbox: Mailboxes iterator.
        """
        return cls('/mailboxes.json', session=session)

    @classmethod
    def get_folders(cls, session, mailbox_or_id):
        """List the folders for the mailbox.

        Args:
            mailbox_or_id (Mailbox or int): Mailbox of the ID of the mailbox
             to get the folders for.

        Returns:
            RequestPaginator of Folder: Folders iterator.
        """
        if isinstance(mailbox_or_id, Mailbox):
            mailbox_or_id = mailbox_or_id.id
        return cls(
            '/mailboxes/%d/folders.json' % mailbox_or_id,
            session=session,
            out_type=Folder,
        )
